@@ -45,6 +45,7 @@ package org.eclipse.jgit.dircache;
 import java.io.File;
 import java.io.FileOutputStream;
 import java.io.IOException;
+import java.io.OutputStream;
 import java.text.MessageFormat;
 import java.util.ArrayList;
 import java.util.HashMap;
@@ -62,6 +63,7 @@ import org.eclipse.jgit.lib.ObjectId;
 import org.eclipse.jgit.lib.ObjectLoader;
 import org.eclipse.jgit.lib.ObjectReader;
 import org.eclipse.jgit.lib.Repository;
+import org.eclipse.jgit.lib.CoreConfig.AutoCRLF;
 import org.eclipse.jgit.treewalk.AbstractTreeIterator;
 import org.eclipse.jgit.treewalk.CanonicalTreeParser;
 import org.eclipse.jgit.treewalk.EmptyTreeIterator;
@@ -73,6 +75,7 @@ import org.eclipse.jgit.treewalk.WorkingTreeOptions;
 import org.eclipse.jgit.treewalk.filter.PathFilter;
 import org.eclipse.jgit.util.FS;
 import org.eclipse.jgit.util.FileUtils;
+import org.eclipse.jgit.util.io.AutoCRLFOutputStream;
 
 /**
  * This class handles checking out one or two trees merging with the index.
@@ -469,6 +472,23 @@ public class DirCacheCheckout {
 		}
 	}
 
+	/**
+	 * Compares whether two pairs of ObjectId and FileMode are equal.
+	 *
+	 * @param id1
+	 * @param mode1
+	 * @param id2
+	 * @param mode2
+	 * @return <code>true</code> if FileModes and ObjectIds are equal.
+	 *         <code>false</code> otherwise
+	 */
+	private boolean equalIdAndMode(ObjectId id1, FileMode mode1, ObjectId id2,
+			FileMode mode2) {
+		if (!mode1.equals(mode2))
+			return false;
+		return id1 != null ? id1.equals(id2) : id2 == null;
+	}
+
 	/**
 	 * Here the main work is done. This method is called for each existing path
 	 * in head, index and merge. This method decides what to do with the
@@ -505,6 +525,9 @@ public class DirCacheCheckout {
 		ObjectId iId = (i == null ? null : i.getEntryObjectId());
 		ObjectId mId = (m == null ? null : m.getEntryObjectId());
 		ObjectId hId = (h == null ? null : h.getEntryObjectId());
+		FileMode iMode = (i == null ? null : i.getEntryFileMode());
+		FileMode mMode = (m == null ? null : m.getEntryFileMode());
+		FileMode hMode = (h == null ? null : h.getEntryFileMode());
 
 		/**
 		 * <pre>
@@ -593,7 +616,7 @@ public class DirCacheCheckout {
 			case 0xFDD: // 10 11
 				// TODO: make use of tree extension as soon as available in jgit
 				// we would like to do something like
-				// if (!iId.equals(mId))
+				// if (!equalIdAndMode(iId, iMode, mId, mMode)
 				//   conflict(name, i.getDirCacheEntry(), h, m);
 				// But since we don't know the id of a tree in the index we do
 				// nothing here and wait that conflicts between index and merge
@@ -607,7 +630,7 @@ public class DirCacheCheckout {
 				conflict(name, (i != null) ? i.getDirCacheEntry() : null, h, m);
 				break;
 			case 0xFDF: // 7 8 9
-				if (hId.equals(mId)) {
+				if (equalIdAndMode(hId, hMode, mId, mMode)) {
 					if (isModified(name))
 						conflict(name, i.getDirCacheEntry(), h, m); // 8
 					else
@@ -622,7 +645,7 @@ public class DirCacheCheckout {
 				keep(i.getDirCacheEntry());
 				break;
 			case 0xFFD: // 12 13 14
-				if (hId.equals(iId)) {
+				if (equalIdAndMode(hId, hMode, iId, iMode)) {
 					dce = i.getDirCacheEntry();
 					if (f == null || f.isModified(dce, true))
 						conflict(name, dce, h, m);
@@ -659,7 +682,9 @@ public class DirCacheCheckout {
 				if (!FileMode.GITLINK.equals(m.getEntryFileMode())) {
 					// a dirty worktree: the index is empty but we have a
 					// workingtree-file
-					if (mId == null || !mId.equals(f.getEntryObjectId())) {
+					if (mId == null
+							|| !equalIdAndMode(mId, mMode,
+									f.getEntryObjectId(), f.getEntryFileMode())) {
 						conflict(name, null, h, m);
 						return;
 					}
@@ -700,7 +725,7 @@ public class DirCacheCheckout {
 				 * </pre>
 				 */
 
-				if (m == null || mId.equals(iId)) {
+				if (m == null || equalIdAndMode(mId, mMode, iId, iMode)) {
 					if (m==null && walk.isDirectoryFileConflict()) {
 						if (dce != null
 								&& (f == null || f.isModified(dce, true)))
@@ -729,7 +754,7 @@ public class DirCacheCheckout {
 					// be removed from the index, but not deleted from disk.
 					remove(name);
 				} else {
-					if (hId.equals(iId)) {
+					if (equalIdAndMode(hId, hMode, iId, iMode)) {
 						if (f == null || f.isModified(dce, true))
 							conflict(name, dce, h, m);
 						else
@@ -738,9 +763,12 @@ public class DirCacheCheckout {
 						conflict(name, dce, h, m);
 				}
 			} else {
-				if (!hId.equals(mId) && !hId.equals(iId) && !mId.equals(iId))
+				if (!equalIdAndMode(hId, hMode, mId, mMode)
+						&& !equalIdAndMode(hId, hMode, iId, iMode)
+						&& !equalIdAndMode(mId, mMode, iId, iMode))
 					conflict(name, dce, h, m);
-				else if (hId.equals(iId) && !mId.equals(iId)) {
+				else if (equalIdAndMode(hId, hMode, iId, iMode)
+						&& !equalIdAndMode(mId, mMode, iId, iMode)) {
 					// For submodules just update the index with the new SHA-1
 					if (dce != null
 							&& FileMode.GITLINK.equals(dce.getFileMode())) {
@@ -926,14 +954,19 @@ public class DirCacheCheckout {
 		ObjectLoader ol = or.open(entry.getObjectId());
 		File parentDir = f.getParentFile();
 		File tmpFile = File.createTempFile("._" + f.getName(), null, parentDir);
-		FileOutputStream channel = new FileOutputStream(tmpFile);
+		WorkingTreeOptions opt = repo.getConfig().get(WorkingTreeOptions.KEY);
+		FileOutputStream rawChannel = new FileOutputStream(tmpFile);
+		OutputStream channel;
+		if (opt.getAutoCRLF() == AutoCRLF.TRUE)
+			channel = new AutoCRLFOutputStream(rawChannel);
+		else
+			channel = rawChannel;
 		try {
 			ol.copyTo(channel);
 		} finally {
 			channel.close();
 		}
 		FS fs = repo.getFS();
-		WorkingTreeOptions opt = repo.getConfig().get(WorkingTreeOptions.KEY);
 		if (opt.isFileMode() && fs.supportsExecute()) {
 			if (FileMode.EXECUTABLE_FILE.equals(entry.getRawMode())) {
 				if (!fs.canExecute(tmpFile))
@@ -954,6 +987,9 @@ public class DirCacheCheckout {
 			}
 		}
 		entry.setLastModified(f.lastModified());
-		entry.setLength((int) ol.getSize());
+		if (opt.getAutoCRLF() != AutoCRLF.FALSE)
+			entry.setLength(f.length()); // AutoCRLF wants on-disk-size
+		else
+			entry.setLength((int) ol.getSize());
 	}
 }
