@@ -51,6 +51,7 @@ import java.io.File;
 import java.io.FileNotFoundException;
 import java.io.FileOutputStream;
 import java.io.IOException;
+import java.text.MessageFormat;
 import java.util.Collection;
 import java.util.Collections;
 import java.util.HashMap;
@@ -78,12 +79,14 @@ import org.eclipse.jgit.revwalk.RevCommit;
 import org.eclipse.jgit.revwalk.RevObject;
 import org.eclipse.jgit.revwalk.RevTree;
 import org.eclipse.jgit.revwalk.RevWalk;
+import org.eclipse.jgit.storage.file.ReflogEntry;
 import org.eclipse.jgit.storage.file.ReflogReader;
 import org.eclipse.jgit.treewalk.TreeWalk;
 import org.eclipse.jgit.util.FS;
 import org.eclipse.jgit.util.FileUtils;
 import org.eclipse.jgit.util.IO;
 import org.eclipse.jgit.util.RawParseUtils;
+import org.eclipse.jgit.util.io.SafeBufferedOutputStream;
 
 /**
  * Represents a Git repository.
@@ -494,14 +497,17 @@ public abstract class Repository {
 					if (!Character.isDigit(rev[l]))
 						break;
 				}
-				String distnum = new String(rev, i + 1, l - i - 1);
 				int dist;
-				try {
-					dist = Integer.parseInt(distnum);
-				} catch (NumberFormatException e) {
-					throw new RevisionSyntaxException(
-							JGitText.get().invalidAncestryLength, revstr);
-				}
+				if (l - i > 1) {
+					String distnum = new String(rev, i + 1, l - i - 1);
+					try {
+						dist = Integer.parseInt(distnum);
+					} catch (NumberFormatException e) {
+						throw new RevisionSyntaxException(
+								JGitText.get().invalidAncestryLength, revstr);
+					}
+				} else
+					dist = 1;
 				while (dist > 0) {
 					RevCommit commit = (RevCommit) ref;
 					if (commit.getParentCount() == 0) {
@@ -524,11 +530,15 @@ public abstract class Repository {
 						break;
 					}
 				}
-				if (time != null)
-					throw new RevisionSyntaxException(
-							JGitText.get().reflogsNotYetSupportedByRevisionParser,
-							revstr);
-				i = m - 1;
+				if (time != null) {
+					String refName = new String(rev, 0, i);
+					Ref resolved = getRefDatabase().getRef(refName);
+					if (resolved == null)
+						return null;
+					ref = resolveReflog(rw, resolved, time);
+					i = m;
+				} else
+					i = m - 1;
 				break;
 			case ':': {
 				RevTree tree;
@@ -550,7 +560,7 @@ public abstract class Repository {
 					tree = rw.parseTree(ref);
 				}
 
-				if (i == rev.length - i)
+				if (i == rev.length - 1)
 					return tree.copy();
 
 				TreeWalk tw = TreeWalk.forPath(rw.getObjectReader(),
@@ -610,6 +620,29 @@ public abstract class Repository {
 		return null;
 	}
 
+	private RevCommit resolveReflog(RevWalk rw, Ref ref, String time)
+			throws IOException {
+		int number;
+		try {
+			number = Integer.parseInt(time);
+		} catch (NumberFormatException nfe) {
+			throw new RevisionSyntaxException(MessageFormat.format(
+					JGitText.get().invalidReflogRevision, time));
+		}
+		if (number < 0)
+			throw new RevisionSyntaxException(MessageFormat.format(
+					JGitText.get().invalidReflogRevision, time));
+
+		ReflogReader reader = new ReflogReader(this, ref.getName());
+		ReflogEntry entry = reader.getReverseEntry(number);
+		if (entry == null)
+			throw new RevisionSyntaxException(MessageFormat.format(
+					JGitText.get().reflogEntryNotFound,
+					Integer.valueOf(number), ref.getName()));
+
+		return rw.parseCommit(entry.getNewId());
+	}
+
 	private ObjectId resolveAbbreviation(final String revstr) throws IOException,
 			AmbiguousObjectException {
 		AbbreviatedObjectId id = AbbreviatedObjectId.fromString(revstr);
@@ -1245,7 +1278,7 @@ public abstract class Repository {
 			throws FileNotFoundException, IOException {
 		File headsFile = new File(getDirectory(), filename);
 		if (heads != null) {
-			BufferedOutputStream bos = new BufferedOutputStream(
+			BufferedOutputStream bos = new SafeBufferedOutputStream(
 					new FileOutputStream(headsFile));
 			try {
 				for (ObjectId id : heads) {
