@@ -217,7 +217,7 @@ public class UploadPack {
 
 	private PackWriter.Statistics statistics;
 
-	private UploadPackLogger logger;
+	private UploadPackLogger logger = UploadPackLogger.NULL;
 
 	/**
 	 * Create a new pack upload for an open repository.
@@ -376,6 +376,11 @@ public class UploadPack {
 		this.packConfig = pc;
 	}
 
+	/** @return the configured logger. */
+	public UploadPackLogger getLogger() {
+		return logger;
+	}
+
 	/**
 	 * Set the logger.
 	 *
@@ -1111,7 +1116,7 @@ public class UploadPack {
 		if (sideband)
 			pckOut.end();
 
-		if (logger != null && statistics != null)
+		if (statistics != null)
 			logger.onPackStatistics(statistics);
 	}
 }
